# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

"""Fixtures in order to test the SQLAlchemy backend."""

from __future__ import absolute_import, unicode_literals

import datetime
import os

import fauxfactory
import pytest
from nark.backends.sqlalchemy import objects
from nark.backends.sqlalchemy.storage import SQLAlchemyStore
from nark.items.activity import Activity
from nark.items.category import Category
from nark.items.fact import Fact
from nark.items.tag import Tag
from pytest_factoryboy import register
from sqlalchemy import create_engine, event

from . import common, factories

register(factories.AlchemyCategoryFactory)
register(factories.AlchemyActivityFactory)
register(factories.AlchemyTagFactory)
register(factories.AlchemyFactFactory)


# *** SQLAlchemy fixtures


def implement_transactional_support_fully(engine):
    """
    (lb): A hack to make SQLite SAVEPOINTs work, so that
    session.begin_nested() works, so that if code under
    test calls session.commit(), we can still rollback().
    See the madness explained at:

      http://docs.sqlalchemy.org/en/rel_1_0/dialects/sqlite.html#pysqlite-serializable
    """

    @event.listens_for(engine, "connect")
    def do_connect(dbapi_connection, connection_record):
        # disable pysqlite's emitting of the BEGIN statement entirely.
        # also stops it from emitting COMMIT before any DDL.
        dbapi_connection.isolation_level = None

    @event.listens_for(engine, "begin")
    def do_begin(conn):
        # emit our own BEGIN
        conn.execute("BEGIN")


# (lb): Use 'module' scope so that the Session is only configured once
# for all tests. (Note: I also tried `scope='session'`, which seemed to
# work similarly.)
@pytest.fixture(scope='module')
def alchemy_runner(request):
    """
    Bind an in-memory mock-db to the session object.

    The session object, common.Session, is a global that this module-scoped
    function sets up just once. And then for each test, we'll create a clean
    session to use for testing that we clean up after.

    This is pretty much straight from the factoryboi docs:

      https://factoryboy.readthedocs.io/en/latest/orms.html#sqlalchemy

    """
    engine = create_engine('sqlite:///:memory:')
    implement_transactional_support_fully(engine)
    objects.metadata.bind = engine
    objects.metadata.create_all(engine)
    common.Session.configure(bind=engine)


@pytest.fixture
def alchemy_session(request):
    """
    Create a new session for each test.

    Reset the database and remove it after each test.

    As suggested by the docs:

        [Using factory_boy with ORMs](
            http://factoryboy.readthedocs.org/en/latest/orms.html#sqlalchemy
        )
    """
    # (lb): 2018-08-22: If we don't create a new Session() for each test,
    # there's one test that triggers a warning, TestFactManager.test_get_all:
    #   /home/user/.virtualenvs/dob/lib/python3.5/site-packages/
    #     sqlalchemy/orm/scoping.py:102: SAWarning: At least one scoped session
    #     is already present.  configure() can not affect sessions that have
    #     already been created.
    # Prepare a new, clean session for each test, lest SQLAlchemy complain.
    session = common.Session()

    # Rather than mock session.commit, use a nested transaction, otherwise
    # items' PK attrs will not be updated.
    # (lb): NOTE: I tried using subtransactions but those didn't work (albeit
    #   that was before I added the implement_transactional_support_fully hack).
    #       session.begin(subtransactions=True)  # (lb): Didn't work for me.
    # This works, but only after having hacked pysqlite to emit BEGIN ourselves.
    session.begin_nested()

    def fin():
        # "Rollback the session => no changes to the database."
        session.rollback()
        # (lb): Call rollback() twice. (For some reason one test,
        # TestFactManager.test_get_all, has five of each item in
        # the db after it runs, and after we can rollback once.
        # Which to me doesn't make sense. But whatever I'm only
        # human, I don't need to understand _everything_.)
        #
        # You can see for yourself with a conditional breakpoint, e.g.,
        #
        #   if session.query('name from categories;').all(): import pdb;pdb.set_trace()
        #
        # (lb): I'm not sure that we need to close, but so far hasn't hurt.
        session.close()
        # "Remove it, so that the next test gets a new Session()."
        common.Session.remove()

    request.addfinalizer(fin)

    return session


@pytest.fixture(params=[
    fauxfactory.gen_utf8(),
    fauxfactory.gen_alphanumeric(),
    ':memory:',
])
def db_path_parametrized(request, tmpdir):
    """Parametrized database paths."""
    if request.param == ':memory:':
        path = request.param
    else:
        path = os.path.join(tmpdir.strpath, request.param)
    return path


@pytest.fixture
def alchemy_config(request, base_config):
    """Provide a config that is suitable for sqlalchemy stores."""
    config = base_config.copy()
    config.update({
        'store': 'sqlalchemy',
        'db_engine': 'sqlite',
        'db_path': ':memory:',
    })
    return config


@pytest.fixture(
    params=(
        # SQLite
        {
            'db_engine': 'sqlite',
            'db_path': ':memory:',
        },
        # Non-SQLite
        {
            'db_engine': 'postgres',
            'db_host': fauxfactory.gen_ipaddr(),
            'db_port': fauxfactory.gen_integer(),
            'db_name': fauxfactory.gen_utf8(),
            'db_user': fauxfactory.gen_utf8(),
            'db_password': fauxfactory.gen_utf8(),
        },
        {
            'db_engine': 'postgres',
            'db_host': fauxfactory.gen_ipaddr(),
            'db_name': fauxfactory.gen_utf8(),
            'db_user': fauxfactory.gen_utf8(),
            'db_password': fauxfactory.gen_utf8(),
        },
    ),
)
def alchemy_config_parametrized(request, alchemy_config):
    """
    Provide a parametrized config that is suitable for sqlalchemy stores.

    We need to build our expectation dynamically if we want to use faked config values.
    """
    config = alchemy_config.copy()
    config.update(request.param)
    if request.param['db_engine'] == 'sqlite':
        expectation = '{engine}:///{path}'.format(
            engine=request.param['db_engine'], path=request.param['db_path'])
    else:
        port = request.param.get('db_port', '')
        if port:
            port = ':{}'.format(port)
        expectation = '{engine}://{user}:{password}@{host}{port}/{name}'.format(
            engine=request.param['db_engine'], user=request.param['db_user'],
            password=request.param['db_password'], host=request.param['db_host'],
            port=port, name=request.param['db_name'])
    return (config, expectation)


@pytest.fixture(params=(
    {'db_engine': None,
     'db_path': None},
    # sqlite
    {'db_engine': 'sqlite',
     'db_path': None},
    {'db_engine': 'sqlite',
     'db_path': ''},
    # Non-sqlite
    {'db_engine': 'postgres',
     'db_host': None,
     'db_name': fauxfactory.gen_utf8(),
     'db_user': fauxfactory.gen_utf8(),
     'db_password': fauxfactory.gen_alphanumeric()},
    {'db_engine': 'postgres',
     'db_host': fauxfactory.gen_ipaddr(),
     'db_name': '',
     'db_user': fauxfactory.gen_utf8(),
     'db_password': fauxfactory.gen_alphanumeric()},
    {'db_engine': 'postgres',
     'db_host': fauxfactory.gen_ipaddr(),
     'db_name': fauxfactory.gen_utf8(),
     'db_user': '',
     'db_password': fauxfactory.gen_alphanumeric()},
    {'db_engine': 'postgres',
     'db_host': fauxfactory.gen_ipaddr(),
     'db_name': fauxfactory.gen_utf8(),
     'db_user': fauxfactory.gen_utf8(),
     'db_password': ''}
))
def alchemy_config_missing_store_config_parametrized(request, alchemy_config):
    """
    Provide an alchemy config containing invalid key/value pairs for
    store initialization.
    """
    config = alchemy_config.copy()
    config.update(request.param)
    return config


@pytest.fixture
def alchemy_store(request, alchemy_config, alchemy_runner, alchemy_session):
    """
    Provide a SQLAlchemyStore that uses our test-session.

    Note:
        The engine created as part of the store.__init__() goes simply unused.
    """
    # (lb): There was a note from hamster-lib:
    #   "probably want this to autouse=True"
    # but after reviewing the docs, I'm guessing there's no point, because an
    # autouse fixture does not return a value, and the code we'd be replacing
    # -- alchemy_session -- creates the common.Session() object that each test
    # uses. So if we removed alchemy_session from the parameters for each test,
    # each test would still have to find the session object somehow.
    #   https://docs.pytest.org/en/latest/fixture.html#autouse-fixtures-xunit-setup-on-steroids
    store = SQLAlchemyStore(alchemy_config)
    store.standup(alchemy_session)
    return store


@pytest.fixture
def alchemy_activity_deleted(alchemy_activity_factory):
    alchemy_activity = alchemy_activity_factory()
    alchemy_activity.deleted = True
    return alchemy_activity


# Instance sets
# Convenience fixtures that provide multitudes of certain alchemy instances.
@pytest.fixture
def set_of_categories(alchemy_category_factory):
    """Provide a number of perstent facts at once."""
    return [alchemy_category_factory() for i in range(5)]


@pytest.fixture
def set_of_tags(alchemy_tag_factory):
    """Provide a number of perstent facts at once."""
    return [alchemy_tag_factory() for i in range(5)]


@pytest.fixture
def set_of_alchemy_facts(start_datetime, alchemy_fact_factory):
    """
    Provide a multitude of generic persistent facts.

    Facts have one day offset from each other and last 20 minutes each.
    """
    start = start_datetime
    result = []
    for i in range(5):
        end = start + datetime.timedelta(minutes=20)
        fact = alchemy_fact_factory(start=start, end=end)
        result.append(fact)
        start = start + datetime.timedelta(days=1)
    return result


# Fallback nark object and factory fixtures. Unless we know how factories
# interact.
@pytest.fixture
def category_factory(request, name):
    """Provide a ``nark.Category`` factory."""
    def generate():
        return Category(name, None)
    return generate


@pytest.fixture
def category(request, category_factory):
    """Provide a randomized ``nark.Category`` instance."""
    return category_factory()


@pytest.fixture
def tag_factory(request, name):
    """Provide a ``nark.Tag`` factory."""
    def generate():
        return Tag(name, None)
    return generate


@pytest.fixture
def tag(request, tag_factory):
    """Provide a randomized ``nark.Tag`` instance."""
    return tag_factory()


@pytest.fixture
def activity_factory(request, name, category_factory):
    """
    Provide a ``nark.Activity`` factory.

    Note:
        * The returned activity will have a *new* category associated as well.
        * Values are randomized but *not parametrized*.
    """
    def generate():
        category = category_factory()
        return Activity(name, pk=None, category=category, deleted=False)
    return generate


@pytest.fixture
def activity(request, activity_factory):
    """Provide a randomized ``nark.Activity`` instance."""
    return activity_factory()


@pytest.fixture
def fact_factory(
    request, activity_factory, tag_factory, start_end_datetimes, description,
):
    """
    Provide a ``nark.Fact`` factory.

    Note:
        * The returned fact will have a *new* activity (and by consequence category)
          associated as well.
        * Values are randomized but *not parametrized*.
    """
    def generate():
        activity = activity_factory()
        tags = set([tag_factory() for i in range(1)])
        start, end = start_end_datetimes
        fact = Fact(activity, start, end, pk=None, description=description, tags=tags)
        return fact
    return generate


@pytest.fixture
def fact(request, fact_factory):
    """Return a randomized ``nark.Fact`` instance."""
    return fact_factory()




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides nark raw fact parsing-related functions."""

from __future__ import absolute_import, unicode_literals

from six import text_type

__all__ = (
    'comma_or_join',
)


def comma_or_join(seq):
    # Need os.linesep or is \n fine?
    seq = [part.replace('\n', '\\n') for part in seq]
    if len(seq) > 1:
        sep_str = '{} or {}'.format(', '.join(seq[:-1]), seq[-1])
    else:
        sep_str = seq[0]
    return sep_str


def format_value_truncate(val, trunc_width=None):
    if not val:
        return val
    # (lb): First attempt was to only short first line:
    #   vals = val.splitlines()
    #   if len(vals) > 1:
    #       val = vals[0] + '...'
    # but replacing newlines with representation of same
    # is more meaningful, and less confusing to end user.
    val = '\\n'.join(text_type(val).splitlines())
    if trunc_width is not None:
        if len(val) > trunc_width and trunc_width >= 0:
            val = val[:trunc_width - 3] + '...'
    return val




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

import gettext
import importlib
import sys
from collections import namedtuple

from future.utils import python_2_unicode_compatible

from .helpers import logging as logging_helpers
from .helpers.dev.profiling import timefunc

BackendRegistryEntry = namedtuple(
    'BackendRegistryEntry', ('verbose_name', 'store_class'),
)


REGISTERED_BACKENDS = {
    'sqlalchemy': BackendRegistryEntry(
        'SQLAlchemy',
        'nark.backends.sqlalchemy.SQLAlchemyStore',
    ),
}


# See: https://wiki.python.org/moin/PortingToPy3k/BilingualQuickRef#gettext
# [FIXME]
# Is this correct?
#   http://www.wefearchange.org/2012/06/the-right-way-to-internationalize-your.html
# seems to use ``sys.version_info.major > 3``
kwargs = {}
if sys.version_info.major < 3:
    kwargs['unicode'] = True
gettext.install('nark', **kwargs)


@python_2_unicode_compatible
class NarkControl(object):
    """
    All mandatory config options are set as part of the controller setup.
    Any client may overwrite those values. but we can always asume that the
    controller does have a value set.

    We will try hard to get through with at least always returning the object.
    We should be able to change only the internal service code to then
    decompose it into its required weired format.

    We were compelled to make attach CRUD-methods to our activity, category
    and fact objects. But as all of those depend on access to the store
    anyway they seem to be best be placed here as a central hub.

    Generic CRUD-actions is to be delegated to our store. The Controller itself
    provides general timetracking functions so that our clients do not have to.
    """

    def __init__(self, config):
        self.config = config
        self.lib_logger = self._get_logger()
        # Profiling: _get_store(): Observed: ~ 0.136 to 0.240 secs.
        self.store = self._get_store()
        self.sql_logger = self._sql_logger()
        self.for_your_convenience()

    def standup_store(self):
        created_fresh = self.store.standup()
        return created_fresh

    def for_your_convenience(self):
        self.migrations = self.store.migrations
        self.categories = self.store.categories
        self.activities = self.store.activities
        self.tags = self.store.tags
        self.facts = self.store.facts

    def update_config(self, config):
        """Use a new config dictionary and apply its settings."""
        self.config = config
        self.store = self._get_store()

    @timefunc
    def _get_store(self):
        """
        Setup the store used by this controller.

        This method is in charge off figuring out the store type, its
        instantiation as well as all additional configuration.
        """

        backend = REGISTERED_BACKENDS.get(self.config['store'])
        if not backend:
            raise KeyError(_("No or invalid storage specified."))
        import_path, storeclass = tuple(backend.store_class.rsplit('.', 1))
        # Profiling: importlib.import_module: ~ 0.265 secs.
        backend_module = importlib.import_module(import_path)
        cls = getattr(backend_module, storeclass)
        store = cls(self.config)
        return store

    def _get_logger(self):
        """
        Setup and configure the main logger.

        As the docs suggest we setup just a pseudo handler. Any client that
        actually wants to use logging needs to setup its required handlers
        itself.
        """

        lib_log_level = self.config['lib_log_level']
        lib_logger = logging_helpers.set_logger_level(
            'nark.log', lib_log_level,
        )
        return lib_logger

    def _sql_logger(self):
        """
        Setup and configure the SQLAlchemy database store logger.

        As the docs suggest we setup just a pseudo handler. Any client that
        actually wants to use logging needs to setup its required handlers
        itself.
        """
        return self.store.logger




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

"""Factories providing randomized object instances."""

from __future__ import unicode_literals

import datetime

import factory
import fauxfactory
from future.utils import python_2_unicode_compatible
from nark.items.activity import Activity
from nark.items.category import Category
from nark.items.fact import Fact
from nark.items.tag import Tag


@python_2_unicode_compatible
class CategoryFactory(factory.Factory):
    """Factory providing randomized ``nark.Category`` instances."""

    pk = None
    # Although we do not need to reference to the object beeing created and
    # ``LazyFunction`` seems sufficient it is not as we could not pass on the
    # string encoding. ``LazyAttribute`` allows us to specify a lambda that
    # circumvents this problem.
    name = factory.LazyAttribute(lambda x: fauxfactory.gen_string('utf8'))

    class Meta:
        model = Category


@python_2_unicode_compatible
class ActivityFactory(factory.Factory):
    """Factory providing randomized ``nark.Activity`` instances."""

    pk = None
    name = factory.Faker('word')
    category = factory.SubFactory(CategoryFactory)
    deleted = False

    class Meta:
        model = Activity


@python_2_unicode_compatible
class TagFactory(factory.Factory):
    """Factory providing randomized ``nark.Category`` instances."""

    pk = None
    name = factory.Faker('word')

    class Meta:
        model = Tag


@python_2_unicode_compatible
class FactFactory(factory.Factory):
    """
    Factory providing randomized ``nark.Fact`` instances.

    Instances have a duration of 3 hours.
    """

    pk = None
    activity = factory.SubFactory(ActivityFactory)
    start = factory.Faker('date_time')
    end = factory.LazyAttribute(lambda o: o.start + datetime.timedelta(hours=3))
    description = factory.Faker('paragraph')

    class Meta:
        model = Fact

    @factory.post_generation
    def tags(self, create, extracted, **kwargs):
        """Add new random tags after instance creation."""
        self.tags = set([TagFactory() for i in range(1)])




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

import os

import pytest
from configparser import ConfigParser
from nark.helpers import app_config
from nark.helpers.app_dirs import NarkAppDirs


class TestNarkAppDirs(object):
    """Make sure that our custom AppDirs works as intended."""

    def test_user_data_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch('nark.helpers.app_dirs.appdirs.user_data_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        assert appdir.user_data_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_user_data_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.user_data_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.user_data_dir) is create

    def test_site_data_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch('nark.helpers.app_dirs.appdirs.site_data_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        assert appdir.site_data_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_site_data_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.site_data_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.site_data_dir) is create

    def test_user_config_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch(
            'nark.helpers.app_dirs.appdirs.user_config_dir',
            return_value=path,
        )
        appdir = NarkAppDirs('nark')
        assert appdir.user_config_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_user_config_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.user_config_dir',
                     return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.user_config_dir) is create

    def test_site_config_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch('nark.helpers.app_dirs.appdirs.site_config_dir',
                     return_value=path)
        appdir = NarkAppDirs('nark')
        assert appdir.site_config_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_site_config_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.site_config_dir',
                     return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.site_config_dir) is create

    def test_user_cache_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch('nark.helpers.app_dirs.appdirs.user_cache_dir',
                     return_value=path)
        appdir = NarkAppDirs('nark')
        assert appdir.user_cache_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_user_cache_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.user_cache_dir',
                     return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.user_cache_dir) is create

    def test_user_log_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch('nark.helpers.app_dirs.appdirs.user_log_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        assert appdir.user_log_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_user_log_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.user_log_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.user_log_dir) is create


class TestGetConfigPath(object):
    """Test config pathj retrieval."""

    def test_get_config_path(self, appdirs):
        """Make sure the config target path is constructed to our expectations."""
        expectation = os.path.join(
            appdirs.user_config_dir,
            app_config.DEFAULT_CONFIG_FILENAME,
        )
        result = app_config.get_config_path()
        assert result == expectation


class TestWriteConfigFile(object):
    """Make sure writing a config instance to disk works as expected."""

    def test_file_is_written(self, config_instance, appdirs):
        """
        Make sure the file is written.

        Note: Content is not checked, this is ConfigParsers job.
        """
        app_config.write_config_file(config_instance)
        expected_location = app_config.get_config_path()
        assert os.path.lexists(expected_location)

    def test_return_config_instance(self, config_instance, appdirs):
        """Make sure we return a ``ConfigParser`` instance."""
        result = app_config.write_config_file(config_instance)
        assert isinstance(result, ConfigParser)


class TestLoadConfigFile(object):
    """Make sure file retrival works as expected."""

    def test_no_file_present(self, appdirs, config_instance):
        """
        Make sure we return ``None``.

        Notw:
            We use the ``appdirs`` fixture to make sure the required dirs exist.
        """
        result = app_config.load_config_file(fallback_config_instance=config_instance)
        assert result == config_instance

    def test_file_present(self, config_instance, backend_config):
        """Make sure we try parsing a found config file."""
        result = app_config.load_config_file()
        assert isinstance(result, ConfigParser)


class TestConfigParserToBackendConfig(object):
    """Make sure that conversion works expected."""

    def test_regular_usecase(self, configparser_instance):
        """Make sure basic mechanics work and int/time types are created."""
        cp_instance, expectation = configparser_instance
        result = app_config.configparser_to_backend_config(cp_instance)
        assert result == expectation




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals


class TestAlchemyCategory(object):
    """Make sure our custom methods behave properly."""
    def test_as_hamster(self, alchemy_store, alchemy_category):
        """Make sure that conversion into a ``nark.Category```works as expected."""
        category = alchemy_category.as_hamster(alchemy_store)
        assert category.equal_fields(alchemy_category)


class TestAlchemyActivity(object):
    """Make sure our custom methods behave properly."""
    def test_as_hamster(self, alchemy_store, alchemy_activity):
        """Make sure that conversion into a ``nark.Activity```works as expected."""
        activity = alchemy_activity.as_hamster(alchemy_store)
        assert activity.equal_fields(alchemy_activity)

    def test_activity_has_facts(self, alchemy_store, alchemy_fact_factory):
        """Make sure that an activity can access ``Fact`` instances."""
        alchemy_fact = alchemy_fact_factory()
        assert alchemy_fact.activity
        activity = alchemy_fact.activity
        assert activity.facts


class TestAlchemyTag(object):
    """Make sure our custom methods behave properly."""
    def test_as_hamster(self, alchemy_store, alchemy_tag):
        """Make sure that conversion into a ``nark.Tag``works as expected."""
        tag = alchemy_tag.as_hamster(alchemy_store)
        assert tag.equal_fields(alchemy_tag)


class TestAlchemyFact(object):
    """Make sure our custom methods behave properly."""

    def test_adding_tags(self, alchemy_store, alchemy_fact, alchemy_tag):
        """
        Make sure that adding tags works as expected.

        This is closer to testing we got SQLAlchemy right than actual code.
        """
        assert len(alchemy_fact.tags) == 4
        alchemy_fact.tags.append(alchemy_tag)
        assert len(alchemy_fact.tags) == 5
        assert alchemy_tag in alchemy_fact.tags

    def test_setting_tags(self, alchemy_store, alchemy_fact, alchemy_tag_factory):
        """
        Make sure that adding tags works as expected.

        This is closer to testing we got SQLAlchemy right than actual code.
        """
        assert alchemy_fact.tags
        new_tags = [alchemy_tag_factory() for i in range(5)]
        alchemy_fact.tags = new_tags
        assert len(alchemy_fact.tags) == 5
        assert alchemy_fact.tags == new_tags

    def test_as_hamster(self, alchemy_store, alchemy_fact):
        """Make sure that conversion into a ``nark.Fact```works as expected."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        assert fact.equal_fields(alchemy_fact)




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

import logging

import pytest
from nark.manager import BaseStore


class TestController:
    @pytest.mark.parametrize('storetype', ['sqlalchemy'])
    def test_get_store_valid(self, controller, storetype):
        """Make sure we recieve a valid ``store`` instance."""
        # [TODO]
        # Once we got backend registration up and running this should be
        # improved to check actual store type for that backend.
        controller.config['store'] = storetype
        assert isinstance(controller._get_store(), BaseStore)

    def test_get_store_invalid(self, controller):
        """Make sure we get an exception if store retrieval fails."""
        controller.config['store'] = None
        with pytest.raises(KeyError):
            controller._get_store()

    def test_update_config(self, controller, base_config, mocker):
        """Make sure we assign new config and get a new store."""
        controller._get_store = mocker.MagicMock()
        controller.update_config({})
        assert controller.config == {}
        assert controller._get_store.called

    def test_get_logger(self, controller):
        """Make sure we recieve a logger that maches our expectations."""
        logger = controller._get_logger()
        assert isinstance(logger, logging.Logger)
        assert logger.name == 'nark.log'
        # [FIXME]
        # assert len(logger.handlers) == 1
        assert isinstance(logger.handlers[0], logging.NullHandler)

    def test_sql_logger(self, controller):
        """Make sure we recieve a logger that maches our expectations."""
        logger = controller._sql_logger()
        assert isinstance(logger, logging.Logger)
        assert logger.name == 'nark.store'
        assert isinstance(logger.handlers[0], logging.NullHandler)




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.
"""A Fact formatting module, to keep Item.Fact() more pure."""

from __future__ import absolute_import, unicode_literals

from six import text_type

from ..helpers.colored import attr, colorize, set_coloring
from ..helpers.strings import format_value_truncate

__all__ = (
    'friendly_str',
    'html_notif',
    'tags_inline',
    'tags_tuples',
    # Private:
    #  '_actegory_string',
    #  '_description_string',
    #  '_ordered_tagnames',
)


def friendly_str(
    fact,
    shellify=False,
    description_sep=': ',
    tags_sep=': ',
    localize=False,
    include_id=False,
    colorful=False,
    cut_width=None,
    show_elapsed=False,
    omit_empty_actegory=False,
):
    """
    Flexible Fact serializer.
    """
    def _friendly_str(fact):
        was_coloring = set_coloring(colorful)
        # MAYBE/2019-01-28: Truncate meta per cut_width or similar.
        meta = assemble_parts(fact)
        result = format_result(fact, meta)

        # (lb): EXPLAIN: Why do we cast here?
        #   (Was this meant to call repr(result) for orig __repr__ implementation?)
        result = text_type(result)

        set_coloring(was_coloring)
        return result

    def assemble_parts(fact):
        parts = [
            get_id_string(fact),
            get_times_string(fact),
            _actegory_string(fact, shellify, omit_empty_actegory),
        ]
        parts_str = ' '.join(list(filter(None, parts)))
        tags = get_tags_string(fact)
        parts_str += tags_sep + tags if tags else ''
        parts_str += _(" [del]") if fact.deleted else ''
        return parts_str

    def format_result(fact, meta):
        result = '{fact_meta}{description}'.format(
            fact_meta=meta,
            description=_description_string(fact, cut_width, description_sep),
        )
        return result

    def get_id_string(fact):
        if not include_id:
            return ''
        # Format the 🏭 🆔 width to be consistent. Assume lifetime of facts?
        # - 999,999 facts over 100 years would be ~27 facts per day.
        return colorize(
            '(🏭 {})'.format(fact.pk and '{:6d}'.format(fact.pk) or 'None'),
            'grey_78',
        )

    def get_times_string(fact):
        times = ''
        times += get_times_string_start(fact)
        times += get_times_string_end(fact, times)
        times += get_times_duration(fact)
        return times

    def get_times_string_start(fact):
        if not fact.start:
            return ''
        if not fact.localize:
            start_time = fact.start_fmt_utc
        else:
            start_time = fact.start_fmt_local
        start_time = colorize(start_time, 'sandy_brown')
        return start_time

    def get_times_string_end(fact, times):
        # NOTE: The CLI's DATE_TO_DATE_SEPARATORS[0] is 'to'.
        prefix = colorize(' to ', 'grey_85') if times else ''
        if not fact.end:
            if not times:
                end_time = ''
            else:
                # (lb): What's a good term here? '<ongoing>'? Or just 'now'?
                end_time = _('<now>')
        elif not fact.localize:
            end_time = fact.end_fmt_utc
        else:
            end_time = fact.end_fmt_local
        end_time = colorize(end_time, 'sandy_brown')
        return prefix + end_time

    def get_times_duration(fact):
        if not show_elapsed:
            return ''
        duration = ' [{}]'.format(fact.format_delta(style=''))
        return colorize(duration, 'grey_78')

    def get_tags_string(fact):
        # (lb): There are three ways to "shellify" a hashtag token:
        #         1.) "#quote" it;
        #         2.) \#delimit it; or
        #         3.) use the inoffensive @ symbol instead of #.
        # Let's do 1.) by default, because most people associate the pound
        # sign with tags, because quotes are less offensive than a slash,
        # and because the @ symbol makes me think of "at'ing someone".
        #   Nope:  hashtag_token = '@' if shellify else '#'
        return fact.tags_inline(quote_tokens=shellify, underlined=colorful)

    # ***

    return _friendly_str(fact)


# ***

def html_notif(fact):
    """
    A briefer Fact one-liner using HTML. Useful for, e.g., notifier toast.
    """
    was_coloring = set_coloring(False)
    duration = '[{}]'.format(fact.format_delta(style=''))
    actegory = _actegory_string(fact, omit_empty_actegory=True)
    actegory = actegory or '<i>No activity</i>'
    description = _description_string(fact, cut_width=39, sep=': ')
    simple_str = (
        '{} {}{}'
        .format(
            duration,
            actegory,
            description,
        )
    )
    set_coloring(was_coloring)
    return simple_str


# ***

def _actegory_string(fact, shellify=False, omit_empty_actegory=False):
    # (lb): We can skip delimiter after time when using ISO 8601.
    if not fact.activity_name:
        if not fact.category_name:
            act_cat = '' if omit_empty_actegory else '@'
        else:
            act_cat = '@{}'.format(fact.category_name)
    else:
        act_cat = (
            '{}@{}'.format(
                fact.activity_name,
                fact.category_name,
            )
        )
    # FIXME: Skinify these colors.
    act_cat = colorize(act_cat, 'cornflower_blue', 'bold', 'underlined')
    act_cat = '"{}"'.format(act_cat) if act_cat and shellify else act_cat
    return act_cat


# ***

def _description_string(fact, cut_width=None, sep=', '):
    description = fact.description or ''
    if description:
        if cut_width is not None:
            description = format_value_truncate(description, cut_width)
        description = '{}{}'.format(sep, description)
    return description


# ***

# (lb): People associate tags with pound signs -- like, #hashtag!
# But Bash, and other shells, use octothorpes to start comments.
# The user can tell Bash to interpret a pound sign as input by
# "#quoting" it, or by \#delimiting it. Hamster also lets the user
# use an '@' at symbol instead (not to be confused with typical
# social media usage of '@' to refer to other users or people).
# By default, this function assumes the tags do not need special
# delimiting; that the pound sign is fine.
def tags_inline(
    fact,
    hashtag_token='#',
    quote_tokens=False,
    colorful=False,
    underlined=False,
):
    def format_tagname(tag):
        tagged = '{}{}'.format(
            # FIXME: Skinify these colors.
            colorize(hashtag_token, 'grey_78') if colorful else hashtag_token,
            colorize(tag.name, 'dark_olive_green_1b') if colorful else tag.name,
        )
        # FIXME: Skinify underlined.
        if underlined:
            tagged = '{}{}{}'.format(
                attr('underlined'), tagged, attr('res_underlined'),
            )
        if quote_tokens:
            tagged = '"{}"'.format(tagged)
        return tagged

    # NOTE: The returned string includes leading space if nonempty!
    tagnames = ''
    if fact.tags:
        tagnames = ' '.join(_ordered_tagnames(fact, format_tagname))
    return tagnames


def tags_tuples(
    fact,
    hashtag_token='#',
    quote_tokens=False,
    colorful=False,
    underlined=False,
    split_lines=False,
):
    def format_tagname(tag):
        # FIXME: Skinify underlined.
        underline_fmt = ' underline' if underlined else ''
        tagged = []
        #
        # FIXME: Skinify these colors.
        token_fmt = 'fg: ' if (colorful or underline_fmt) else ''
        token_fmt += '#C6C6C6' if colorful else ''
        token_fmt += underline_fmt
        tagged.append((token_fmt, hashtag_token))
        #
        tname_fmt = 'fg: ' if (colorful or underline_fmt) else ''
        tname_fmt += '#D7FF87' if colorful else ''
        tname_fmt += underline_fmt
        tagged.append((tname_fmt, tag.name))
        #
        if quote_tokens:
            fmt_quote = ('', '"')
            tagged.insert(0, fmt_quote)
            tagged.append(fmt_quote)
        return tagged

    # NOTE: The returned string includes leading space if nonempty!
    tagnames = []
    if fact.tags:
        fmt_sep = ('', "\n") if split_lines else ('', ' ')
        n_tag = 0
        for fmtd_tagn in _ordered_tagnames(fact, format_tagname):
            if n_tag > 0:
                tagnames += [fmt_sep]
            n_tag += 1
            tagnames += fmtd_tagn
    return tagnames


def _ordered_tagnames(fact, format_tagname):
    return [
        format_tagname(tag) for tag in fact.tags_sorted
    ]




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides nark raw fact parsing-related functions."""

from __future__ import absolute_import, unicode_literals

__all__ = (
    'ParserException',
    'ParserMissingDatetimeException',
    'ParserMissingDatetimeOneException',
    'ParserMissingDatetimeTwoException',
    'ParserInvalidDatetimeException',
    'ParserMissingSeparatorActivity',
    'ParserMissingActivityException',
)


class ParserException(Exception):
    """Raised if parser cannot decipher nark factoid string."""
    pass


class ParserMissingDatetimeException(ParserException):  # noqa: E302
    """Raised if the raw_fact is missing one or both datetime tokens."""
    pass


class ParserMissingDatetimeOneException(ParserMissingDatetimeException):  # noqa: E302
    """Raised if the raw_fact is missing its start datetime token(s)."""
    pass


class ParserMissingDatetimeTwoException(ParserMissingDatetimeException):  # noqa: E302
    """Raised if the raw_fact is missing its end datetime token(s)."""
    pass


class ParserInvalidDatetimeException(ParserException):  # noqa: E302
    """Raised if a time from raw_fact in not parseworthy."""
    pass


class ParserMissingSeparatorActivity(ParserException):  # noqa: E302
    """Raised if activity@category separator not found."""
    pass


class ParserMissingActivityException(ParserException):  # noqa: E302
    """Raised if factoid is missing: act@cat, cat@, @cat, or just @."""
    pass




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

import os

import appdirs

__all__ = (
    'NarkAppDirs',
)


class NarkAppDirs(appdirs.AppDirs):
    """Custom class that ensure appdirs exist."""

    def __init__(self, *args, **kwargs):
        """Add create flag value to instance."""
        super(NarkAppDirs, self).__init__(*args, **kwargs)
        # FIXME: (lb): I'm not super cool with this side-effect:
        #          Calling any property herein will cause its
        #          directory path to be created! Creating paths
        #          should be a deliberate action and not a side effect
        #          of just asking for a path. In any case, it currently
        #          works this way, so just rolling with the flow, for now.
        #        See Click: it has concept of lazy-creating paths, i.e.,
        #          only create path when a file therein opened for write.
        self.create = True

    @property
    def user_data_dir(self):
        """Return ``user_data_dir``."""
        directory = appdirs.user_data_dir(
            self.appname,
            self.appauthor,
            version=self.version,
            roaming=self.roaming,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    @property
    def site_data_dir(self):
        """Return ``site_data_dir``."""
        directory = appdirs.site_data_dir(
            self.appname,
            self.appauthor,
            version=self.version,
            multipath=self.multipath,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    @property
    def user_config_dir(self):
        """Return ``user_config_dir``."""
        directory = appdirs.user_config_dir(
            self.appname,
            self.appauthor,
            version=self.version,
            roaming=self.roaming,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    @property
    def site_config_dir(self):
        """Return ``site_config_dir``."""
        directory = appdirs.site_config_dir(
            self.appname,
            self.appauthor,
            version=self.version,
            multipath=self.multipath,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    @property
    def user_cache_dir(self):
        """Return ``user_cache_dir``."""
        directory = appdirs.user_cache_dir(
            self.appname, self.appauthor, version=self.version,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    @property
    def user_log_dir(self):
        """Return ``user_log_dir``."""
        directory = appdirs.user_log_dir(
            self.appname, self.appauthor, version=self.version,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    def _ensure_directory_exists(self, directory):
        """Ensure that the passed path exists."""
        if not os.path.lexists(directory):
            os.makedirs(directory)
        return directory




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

"""nark provides generic time tracking functionality."""

from __future__ import absolute_import, unicode_literals

import os
import sys
import time

__all__ = (
    '__package_name__',
    '__resolve_vers__',
    '__time_0__',
    '__PROFILING__',
)

__PROFILING__ = True
# DEVS: Comment this out to see load times summary.
__PROFILING__ = False
__time_0__ = time.time()

# (lb): Seems a little redundant (see setup.cfg:[metadata]name)
# but not sure if way to get programmatically. This is closest
# solution that avoids hardcoding the library name in strings
# (which is something linter or runtime won't catch if wrong).
__package_name__ = 'nark'


def __resolve_vers__():
    """Returns the installed package version, or '<none>'.

    In lieu of always setting __version__ -- and always loading pkg_resources --
    use a method to avoid incurring startup costs if the version is not needed.
    """
    try:
        import setuptools_scm
        # For whatever reason, relative_to does not work, (lb) thought it would.
        #   return setuptools_scm.get_version(relative_to=__file__)
        pkg_dirname = os.path.dirname(os.path.dirname(__file__))
        # See if parent directory (of this file) contains .git/.
        proj_git = os.path.join(pkg_dirname, '.git')
        if os.path.exists(proj_git):
            # Get version from setuptools_scm, and git tags.
            # This is similar to a developer running, e.g.,
            #   python setup.py --version
            return setuptools_scm.get_version(root=pkg_dirname)
    except ImportError:
        # ModuleNotFoundError added py3.6. Until then, ImportError.
        from pkg_resources import get_distribution, DistributionNotFound
        try:
            return get_distribution(__package_name__).version
        except DistributionNotFound:
            return '<none>'
    else:
        return '<none>'




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

from collections import namedtuple
from datetime import datetime
from operator import attrgetter

from future.utils import python_2_unicode_compatible
from six import text_type

from ..helpers import fact_time, format_fact, format_time
from ..helpers.parsing import parse_factoid
from .activity import Activity
from .category import Category
from .item_base import BaseItem
from .tag import Tag

try:
    from math import inf
except ImportError:
    # < Python 3.5.
    inf = float('inf')


__all__ = (
    'SinceTimeBegan',
    'UntilTimeStops',
    'FactTuple',
    'Fact',
)


SinceTimeBegan = datetime(1, 1, 1)


UntilTimeStops = datetime(9999, 12, 31, 23, 59, 59)


FactTuple = namedtuple(
    'FactTuple',
    (
        'pk',
        'activity',
        'start',
        'end',
        'description',
        'tags',
        'deleted',
        'split_from',
    ),
)


@python_2_unicode_compatible
class Fact(BaseItem):
    """Storage agnostic class for facts."""
    def __init__(
        self,
        activity,
        start,
        end=None,
        pk=None,
        description=None,
        tags=None,
        deleted=False,
        split_from=None,
    ):
        """
        Initiate our new instance.

        Args:
            activity (nark.Activity): Activity associated with this fact.

            start (datetime.datetime): Start datetime of this fact.

            end (datetime.datetime, optional): End datetime of this fact.
                Defaults to ``None``.

            pk (optional): Primary key used by the backend to identify this instance.
                Defaults to ``None``.

            description (str, optional): Additional information relevant to this
                singular fact. Defaults to ``None``.

            tags (Iterable, optional): Iterable of ``strings`` identifying *tags*.
                Defaults to ``None``.

            deleted (bool, optional): True if fact was deleted/edited/split.

            split_from (nark.Fact.id, optional): ID of deleted fact this
                fact succeeds.
        """
        super(Fact, self).__init__(pk, name=None)
        assert activity is None or isinstance(activity, Activity)
        self.activity = activity
        self.start = start
        self.end = end
        self.description = description

        self.tags_replace(tags)

        # (lb): Legacy Hamster did not really have an edit-fact feature.
        # Rather, when the user "edited" a Fact, Hamster would delete
        # the existing row and make a new one. This is very unwikilike!
        # To preserve history, let's instead mark edited Facts deleted.
        # FIXME/2018-05-23 10:56: (lb): Add column to record new Facts ID?
        self.deleted = bool(deleted)

        self.split_from = split_from

    def __eq__(self, other):
        if other is not None and not isinstance(other, FactTuple):
            other = other.as_tuple()

        return self.as_tuple() == other

    def equal_sans_end(self, other):
        if other is not None and not isinstance(other, FactTuple):
            other = other.as_tuple(sans_end=True)

        return self.as_tuple(sans_end=True) == other

    def __hash__(self):
        """Naive hashing method."""
        return hash(self.as_tuple())

    def __gt__(self, other):
        return self.sorty_tuple > other.sorty_tuple

    def __lt__(self, other):
        return self.sorty_tuple < other.sorty_tuple

    @property
    def sorty_times(self):
        fact_end = self.end if self.end is not None else UntilTimeStops
        return (self.start, fact_end)

    @property
    def sorty_tuple(self):
        fact_end = self.end if self.end is not None else UntilTimeStops
        fact_pk = self.pk if self.pk is not None else -inf
        return (self.start, fact_end, fact_pk)

    def __str__(self):
        # MAYBE: (lb): __str__ used to pass fmttr=text_type to friendly_str,
        #   and __repr__ used to pass fmttr=repr; and then friendly_str would
        #   wrap all strings that it assembled with that formatter, e.g.,
        #     fmttr(self.activity.name)
        #   but I did not understand why (what's the difference between
        #   text_type and repr? text_type is a no-op in py3, and converts
        #   py2 string to unicode, right?). In any case, I don't think we
        #   need to cast to text_type/repr anymore (especially if we drop
        #   Python 2 support).
        return format_fact.friendly_str(self)

    def __repr__(self):
        parts = []
        for key in sorted(self.__dict__.keys()):
            if key == 'name':
                assert self.name is None
                continue  # Part of BaseItem, and not used by Fact.
            elif key == 'tags':
                val = repr(self.tags_sorted)
            else:
                # Not that nark knows anything about this, but dob's derived
                # class, PlaceableFact, has a next_fact and prev_fact pointer.
                # To avoid a deadly RecursionError by following pointers of a
                # doubly-linked list, check that the attribute value is not
                # another Fact. (For the sake of coupling, PlaceableFact could
                # just supply its own __repr__() function, but this function
                # is already doing a lot, so might as well toss in the kitchen
                # sink.)
                raw = getattr(self, key)
                if isinstance(raw, self.__class__):
                    val = id(raw)
                else:
                    val = repr(raw)
            parts.append(
                "{key}={val}".format(key=key, val=val)
            )
        repred = "Fact({})".format(', '.join(parts))
        return repred

    def as_tuple(self, include_pk=True, sans_end=False):
        """
        Provide a tuple representation of this facts relevant attributes.

        Args:
            include_pk (bool): Whether to include the instances pk or not. Note that if
            ``False`` ``tuple.pk = False``!

        Returns:
            nark.FactTuple: Representing this categories values.
        """
        pk = self.pk
        if not include_pk:
            pk = False

        activity_tup = self.activity and self.activity.as_tuple(include_pk=include_pk)

        sorted_tags = self.tags_sorted
        ordered_tags = [tag.as_tuple(include_pk=include_pk) for tag in sorted_tags]

        end_time = -1 if sans_end else self.end

        return FactTuple(
            pk=pk,
            activity=activity_tup,
            start=self.start,
            end=end_time,
            description=self.description,
            tags=frozenset(ordered_tags),
            deleted=bool(self.deleted),
            split_from=self.split_from,
        )

    def copy(self, include_pk=True):
        """
        """
        new_fact = self.__class__(
            activity=self.activity,
            start=self.start,
            end=self.end,
            description=self.description,
            # self.tags might be an sqlalchemy.orm.collections.InstrumentedList
            # and calling list() on it will create a new list of what could be
            # nark.backends.sqlalchemy.objects.AlchemyTag.
            tags=list(self.tags),
            deleted=bool(self.deleted),
            split_from=self.split_from,
        )
        if include_pk:
            new_fact.pk = self.pk
        return new_fact

    def equal_fields(self, other):
        """
        Compare this instances fields with another fact. This excludes comparing the PK.

        Args:
            other (Fact): Fact to compare this instance with.

        Returns:
            bool: ``True`` if all fields but ``pk`` are equal, ``False`` if not.

        Note:
            This is particularly useful if you want to compare a new ``Fact`` instance
            with a freshly created backend instance. As the latter will probably have a
            primary key assigned now and so ``__eq__`` would fail.
        """
        return self.as_tuple(include_pk=False) == other.as_tuple(include_pk=False)

    # ***

    LOCALIZE = False

    @classmethod
    def localize(cls, localize=None):
        if localize is not None:
            cls.LOCALIZE = localize
        return cls.LOCALIZE

    # ***

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        """
        Make sure that we receive a ``datetime.datetime`` instance,
          or relative time string.

        Args:
            start (datetime.datetime, text_type): Start datetime of this ``Fact``.

        Raises:
            TypeError: If we receive something other than a ``datetime.datetime``
                (sub-)class or ``None``.
        """
        # MOTE: (lb): The AlchemyFact class derives from this class, but it
        # has columns of the same names as theses @property definitions, e.g.,
        # `start`. So when you set, e.g, `self.start = X` from the AlchemyFact
        # class, it does not call this base class' @setter for start. So don't
        # use self._start except in self.start()/=.
        self._start = fact_time.must_be_datetime_or_relative(start)

    @property
    def start_fmt_utc(self):
        """FIXME: Document"""
        if not self.start:
            return ''
        # Format like: '%Y-%m-%d %H:%M:%S%z'
        return format_time.isoformat_tzinfo(self.start, sep=' ', timespec='seconds')

    @property
    def start_fmt_local(self):
        """FIXME: Document"""
        if not self.start:
            return ''
        # Format like: '%Y-%m-%d %H:%M:%S'
        return format_time.isoformat_tzless(self.start, sep=' ', timespec='seconds')

    # ***

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        """
        Make sure that we receive a ``datetime.datetime`` instance.

        Args:
            end (datetime.datetime): End datetime of this ``Fact``.

        Raises:
            TypeError: If we receive something other than a ``datetime.datetime``
                (sub-)class or ``None``.
        """
        self._end = fact_time.must_be_datetime_or_relative(end)

    @property
    def end_fmt_utc(self):
        """FIXME: Document"""
        if not self.end:
            return ''
        return format_time.isoformat_tzinfo(self.end, sep=' ', timespec='seconds')

    @property
    def end_fmt_local(self):
        """FIXME: Document"""
        if not self.end:
            return ''
        return format_time.isoformat_tzless(self.end, sep=' ', timespec='seconds')

    @property
    def end_fmt_local_nowwed(self):
        """FIXME: Document"""
        if not self.end:
            return '{} <now>'.format(self.end_fmt_local_or_now)
        return self.end_fmt_local

    @property
    def end_fmt_local_or_now(self):
        if not self.end:
            return '{}'.format(format_time.isoformat_tzless(
                self.time_now, sep=' ', timespec='seconds',
            ))
        return self.end_fmt_local

    # ***

    @property
    def momentaneous(self):
        if self.times_ok and self.start == self.end:
            return True
        return False

    @property
    def time_now(self):
        return datetime.now() if self.localize else datetime.utcnow()

    @property
    def times(self):
        return (self.start, self.end)

    @property
    def times_ok(self):
        if isinstance(self.start, datetime) and isinstance(self.end, datetime):
            return True
        return False

    # ***

    def delta(self):
        """
        Provide the offset of start to end for this fact.

        Returns:
            datetime.timedelta or None: Difference between start- and end datetime.
                If we only got a start datetime, return ``None``.
        """
        end_time = self.end
        if not end_time:
            end_time = self.time_now

        return end_time - self.start

    def format_delta(self, style='%M'):
        """
        Return a string representation of ``Fact().delta``.

        Args:
            formatting (str): Specifies the output format.

              Valid choices are:
                * ``'%M'``: As minutes, rounded down.
                * ``'%H:%M'``: As 'hours:minutes'. rounded down.
                * ``'HHhMMm'``: As '{hours} hour(s) {minutes} minute(s)'.
                * ``''``: As human friendly time.

        Returns:
            str: Formatted string representing this fact's *duration*.
        """
        return format_time.format_delta(self.delta(), style=style)

    # ***

    @property
    def midpoint(self):
        if not self.times_ok:
            return None
        midpoint = self.end - ((self.end - self.start) / 2)
        return midpoint

    @property
    def time_of_day_midpoint(self):
        if not self.midpoint:
            return ''
        clock_sep = ' ◐ '
        hamned = '{0}'.format(
            self.midpoint.strftime("%a %d %b %Y{0}%I:%M %p").format(clock_sep),
            # FIXME: (lb): Add Colloquial TOD suffix, e.g., "morning".
        )
        return hamned

    def time_of_day_humanize(self, show_now=False):
        if not self.times_ok and not show_now:
            return ''
        clock_sep = ' ◐ '
        wkd_day_mon_year = self.start.strftime("%a %d %b %Y")
        text = self.start.strftime("{0}{1}%I:%M %p").format(
            wkd_day_mon_year, clock_sep,
        )
        if self.end == self.start:
            return text
        text += _(" — ")
        end_time = self.end if self.end is not None else self.time_now
        text += end_time.strftime("%I:%M %p")
        end_wkd_day_mon_year = end_time.strftime("%a %d %b %Y")
        if end_wkd_day_mon_year == wkd_day_mon_year:
            return text
        text += " "
        text += end_wkd_day_mon_year
        return text

    # ***

    @property
    def activity_name(self):
        """..."""
        try:
            return self.activity.name
        except AttributeError:
            return ''

    @property
    def category(self):
        """Just a convenience shim to underlying category object."""
        return self.activity.category

    @property
    def category_name(self):
        """..."""
        try:
            return self.activity.category.name
        except AttributeError:
            return ''

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        """"
        Normalize all descriptions that evaluate to ``False``.
        Store everything else as string.
        """
        if description:
            description = text_type(description)
        else:
            description = None
        self._description = description

    def tags_replace(self, tags):
        new_tags = set()
        for tagn in set(tags) if tags else set():
            tag = tagn if isinstance(tagn, Tag) else Tag(name=tagn)
            new_tags.add(tag)
        # (lb): Do this in one swoop, and be sure to assign a list; when wrapped
        # by SQLAlchemy, if tried to set to, e.g., set(), complains:
        #   TypeError: Incompatible collection type: set is not list-like
        # (That error is from orm.attribute.CollectionAttributeImpl.set.)
        self.tags = list(new_tags)

    @property
    def tags_sorted(self):
        return sorted(list(self.tags), key=attrgetter('name'))

    # ***

    def tagnames(self, *args, **kwargs):
        return format_fact.tags_inline(self, *args, **kwargs)

    def tags_inline(self, colorful=False, underlined=False, **kwargs):
        return format_fact.tags_inline(
            self, colorful=colorful, underlined=underlined, **kwargs
        )

    def tags_tuples(self, colorful=False, underlined=False, **kwargs):
        return format_fact.tags_tuples(
            self, colorful=colorful, underlined=underlined, **kwargs
        )

    # ***

    def friendly_str(self, *args, **kwargs):
        return format_fact.friendly_str(self, *args, **kwargs)

    # ***

    def get_serialized_string(self, shellify=False):
        """
        Return a canonical, "stringified" version of the Fact.

        - Akin to: encoding/flattening/marshalling/packing/pickling.

        This function is mostly meant for machines, not for people.

        - Generally, use ``__str__`` if you want a human-readable string.

          I.e., one whose datetimes are localized relative to the Fact.
          This serializing function defaults to using UTC.

        - Use this function to encode a Fact in a canonical way, which can
          be consumed again later, i.e., using ``Fact.create_from_factoid``.

        - A complete serialized fact might look like this:

              2016-02-01 17:30 to 2016-02-01 18:10 making plans@world domination
              #tag 1 #tag 2, description

          - Note that nark is very unassuming with whitespace. It can be
            used in the Activity and Category names, as well as in tags.

        Attention:

            ``Fact.tags`` is a set and hence unordered. In order to provide
            a deterministic canonical return string, we sort tags by name.
            This is purely cosmetic and does not imply any actual ordering
            of those facts on the instance level.

        Returns:
            text_type: Canonical string encoding all available fact info.
        """
        return format_fact.friendly_str(self, shellify=shellify)

    @property
    def short(self):
        """
        A brief Fact one-liner.

        (lb): Not actually called by any code, but useful for debugging!
        """
        return format_fact.friendly_str(self, include_id=True, cut_width=39)

    @property
    def html_notif(self):
        """
        A briefer Fact one-liner using HTML. Useful for, e.g., notifier toast.
        """
        return format_fact.html_notif(self)

    # ***

    @classmethod
    def create_from_factoid(
        cls,
        factoid,
        time_hint='verify_none',
        separators=None,
        lenient=False,
    ):
        """
        Construct a new ``nark.Fact`` from a string of fact details,
            or factoid.

        NOTE: This naïvely creates a new Fact and does not check against
        other Facts for integrity. It's up to the caller to see if the new
        Fact conflicts with existing Facts presently in the system.

        Args:
            factoid (str): Raw fact to be parsed.

            time_hint (text_type, optional): One of:
                'verify_none': Do not expect to find any time encoded in factoid.
                'verify_both': Expect to find both start and end times.
                'verify_start': Expect to find just one time, which is the start.
                'verify_end': Expect to find just one time, which is the end.
                'verify_then': Time specifies new start; and back-fill interval gap.
                'verify_after': No time spec. Start new Fact at time of previous end.

            lenient (bool, optional): If False, parser raises errors on misssing
                mandatory components (such as time or activity). (Category,
                tags, and description are optional.)

        Returns:
            nark.Fact: New ``Fact`` object constructed from factoid.

        Raises:
            ValueError: If we fail to extract at least ``start`` or ``activity.name``.
            ValueError: If ``end <= start``.
            ParserException: On parser error, one of the many ParserException
                derived classes will be raised.
        """
        parsed_fact, err = parse_factoid(
            factoid,
            time_hint=time_hint,
            separators=separators,
            lenient=lenient,
        )

        new_fact = cls.create_from_parsed_fact(
            parsed_fact, lenient=lenient,
        )

        return new_fact, err

    @classmethod
    def create_from_parsed_fact(
        cls,
        parsed_fact,
        lenient=False,
        **kwargs
    ):
        start = parsed_fact['start']
        end = parsed_fact['end']
        # Verify that start > end, if neither are None or not a datetime.
        start, end = fact_time.must_not_start_after_end((start, end))

        activity = ''
        activity_name = parsed_fact['activity']
        if activity_name:
            activity = Activity(activity_name)
        elif lenient:
            activity = Activity(name='')
        else:
            raise ValueError(_('Unable to extract activity name'))

        category_name = parsed_fact['category']
        if category_name:
            activity.category = Category(category_name)

        description = parsed_fact['description']

        tags = parsed_fact['tags']

        return cls(
            activity,
            start,
            end=end,
            description=description,
            tags=tags,
            **kwargs
        )




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

from collections import namedtuple

from future.utils import python_2_unicode_compatible
from six import text_type

from .item_base import BaseItem

TagTuple = namedtuple(
    'TagTuple', ('pk', 'name', 'deleted', 'hidden'),
)


@python_2_unicode_compatible
class Tag(BaseItem):
    """Storage agnostic class for tags."""

    def __init__(self, name, pk=None, deleted=False, hidden=False):
        """
        Initialize this instance.

        Args:
            name (str): The name of the tag. May contain whitespace!
            pk: The unique primary key used by the backend.
        """

        super(Tag, self).__init__(pk, name)
        self.deleted = bool(deleted)
        self.hidden = bool(hidden)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if not name:
            # Disallow `None` and empty string.
            raise ValueError(_('Tags must be named.'))
        self._name = text_type(name)

    def as_tuple(self, include_pk=True):
        """
        Provide a tuple representation of this tags relevant 'fields'.

        Args:
            include_pk (bool): Whether to include the instances pk or not.
            Note that if ``False`` ``tuple.pk = False``!

        Returns:
            TagTuple: Representing this tags values.
        """
        pk = self.pk
        if not include_pk:
            pk = False
        tag_tup = TagTuple(
            pk=pk,
            name=self.name,
            deleted=bool(self.deleted),
            hidden=bool(self.hidden),
        )
        return tag_tup

    def equal_fields(self, other):
        """
        Compare this instances fields with another tag. This excludes comparing the PK.

        Args:
            other (Tag): Tag to compare this instance with.

        Returns:
            bool: ``True`` if all fields but ``pk`` are equal, ``False`` if not.

        Note:
            This is particularly useful if you want to compare a new ``Tag`` instance
            with a freshly created backend instance. As the latter will probably have a
            primary key assigned now and so ``__eq__`` would fail.
        """
        if other:
            other = other.as_tuple(include_pk=False)
        else:
            other = None

        return self.as_tuple(include_pk=False) == other

    def __eq__(self, other):
        if other is not None and not isinstance(other, TagTuple):
            other = other.as_tuple()
        return self.as_tuple() == other

    def __hash__(self):
        """Naive hashing method."""
        return hash(self.as_tuple())

    def __str__(self):
        return text_type('{name}'.format(name=self.name))




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

from future.utils import python_2_unicode_compatible

from . import BaseManager


@python_2_unicode_compatible
class BaseMigrationsManager(BaseManager):
    """Base class defining the minimal API for a MigrationsManager implementation."""

    def downgrade(self):
        """Downgrade the database according to its migration version."""
        raise NotImplementedError

    def upgrade(self):
        """Upgrade the database according to its migration version."""
        raise NotImplementedError

    def version(self):
        """Returns the migration version of the database indicated by the config."""
        raise NotImplementedError

    def latest_version(self):
        """Returns the latest version of the database as used by the application."""
        raise NotImplementedError

    # ***

    def migrations_config(self):
        """Return sqlalchemy-migrate migrate.cfg as dictionary object."""
        raise NotImplementedError

    def migrations_path(self):
        """Return the path to the migration versions/ base directory."""
        raise NotImplementedError

    def legacy_upgrade_from_hamster_applet(self, db_path):
        """Upgrade legacy SQLite database created by hamster-applet."""
        raise NotImplementedError

    def legacy_upgrade_from_hamster_lib(self):
        """Upgrade legacy SQLite database created by hamster-lib."""
        raise NotImplementedError




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

"""
Base classes for implementing storage backends.

Note:
    * This is propably going to be replaced by a ``ABC``-bases solution.
    * Basic sanity checks could be done here then. This would mean we just need
      to test them once and our actual backends focus on the CRUD implementation.
"""

from __future__ import absolute_import, unicode_literals

from future.utils import python_2_unicode_compatible

from ..items.item_base import BaseItem

__all__ = ('BaseManager', )


@python_2_unicode_compatible
class BaseManager(object):
    """Base class for all object managers."""

    def __init__(self, store):
        self.store = store

    # ***

    def adding_item_must_not_have_pk(self, hamster_item):
        message = _("Adding item: {!r}.".format(hamster_item))
        self.store.logger.debug(message)
        if not hamster_item.pk:
            return
        message = _(
            "The {} item ({!r}) cannot be added because it already has a PK."
            " Perhaps call the ``_update`` method instead".format(
                self.__class__.__name__, hamster_item,
            )
        )
        self.store.logger.error(message)
        raise ValueError(message)

    # ***

    def save(self, item, cls=BaseItem, named=False, **kwargs):
        """
        Save a Nark object instance to user's selected backend.

        Will either ``_add`` or ``_update`` based on item PK.

        Args:
            tag (nark.BaseItem, i.e., Activity/Category/Fact/Tag):
                Nark instance to be saved.

        Returns:
            nark.BaseItem: Saved Nark instance.

        Raises:
            TypeError: If the ``item`` parameter is not a valid ``BaseItem`` instance.
        """

        if not isinstance(item, cls):
            message = _("You need to pass a {} object").format(cls.__name__)
            self.store.logger.debug(message)
            raise TypeError(message)

        # (lb): Not sure this is quite what we want, but Activity has been doing this,
        # and I just made this base class, so now all items will be doing this.
        if named and not item.name:
            raise ValueError(_("You must specify an item name."))

        self.store.logger.debug(_("'{}' has been received.".format(item)))

        # NOTE: Not assuming that PK is an int, i.e., not testing '> 0'.
        if item.pk or item.pk == 0:
            result = self._update(item, **kwargs)
        else:
            # PK is empty string, empty list, None, etc., but not 0.
            result = self._add(item, **kwargs)
        return result




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

import logging

from nark.helpers import logging as logging_helpers


class TestSetupHandler(object):
    def test_get_formatter_basic(self, mocker):
        """Test formatter fetcher."""
        formatter = logging_helpers.formatter_basic()
        # (lb): Is this legit, or a little too _intimate?
        expected = '[%(levelname)s] %(asctime)s %(name)s %(funcName)s: %(message)s'
        assert formatter._fmt == expected

    def test_setup_handler_stream_handler(self, mocker):
        """Test logging setup."""
        stream_handler = logging.StreamHandler()
        formatter = logging_helpers.formatter_basic()
        logger = mocker.MagicMock()
        logging_helpers.setup_handler(stream_handler, formatter, logger)
        logger.addHandler.assert_called_with(stream_handler)




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

from future.utils import python_2_unicode_compatible

from . import BaseManager
from ..items.category import Category


@python_2_unicode_compatible
class BaseCategoryManager(BaseManager):
    """
    Base class defining the minimal API for a CategoryManager implementation.
    """

    def __init__(self, *args, **kwargs):
        super(BaseCategoryManager, self).__init__(*args, **kwargs)

    # ***

    def save(self, category):
        """
        Save a Category to the backend.

        Args:
            category (nark.Category): Category instance to be saved.

        Returns:
            nark.Category: Saved Category

        Raises:
            TypeError: If the ``category`` parameter is not a valid
                ``Category`` instance.
        """
        return super(BaseCategoryManager, self).save(category, Category, named=True)

    # ***

    def get_or_create(self, category):
        """
        Check if we already got a category with that name, if not create one.

        This is a convenience method as it seems sensible to rather implement
        this once in our controller than having every client implementation
        deal with it anew.

        It is worth noting that the lookup completely ignores any PK contained in the
        passed category. This makes this suitable to just create the desired Category
        and pass it along. One way or the other one will end up with a persisted
        db-backed version.

        Args:
            category (nark.Category or None): The categories.

        Returns:
            nark.Category or None: The retrieved or created category. Either way,
                the returned Category will contain all data from the backend, including
                its primary key.
        """

        self.store.logger.debug(_("'{}' has been received.'.".format(category)))
        if category:
            try:
                category = self.get_by_name(category)
            except KeyError:
                category = Category(category)
                category = self._add(category)
        else:
            # We want to allow passing ``category=None``, so we normalize here.
            category = None
        return category

    # ***

    def _add(self, category):
        """
        Add a ``Category`` to our backend.

        Args:
            category (nark.Category): ``Category`` to be added.

        Returns:
            nark.Category: Newly created ``Category`` instance.

        Raises:
            ValueError: When the category name was already present!
                It is supposed to be unique.
            ValueError: If category passed already got an PK.
                Indicating that update would be more appropriate.

        Note:
            * Legacy version stored the proper name as well as a ``lower(name)`` version
            in a dedicated field named ``search_name``.
        """
        raise NotImplementedError

    # ***

    def _update(self, category):
        """
        Update a ``Categories`` values in our backend.

        Args:
            category (nark.Category): Category to be updated.

        Returns:
            nark.Category: The updated Category.

        Raises:
            KeyError: If the ``Category`` can not be found by the backend.
            ValueError: If the ``Category().name`` is already being used by
                another ``Category`` instance.
            ValueError: If category passed does not have a PK.
        """
        raise NotImplementedError

    # ***

    def remove(self, category):
        """
        Remove a category.

        Any ``Activity`` referencing the passed category will be set to
        ``Activity().category=None``.

        Args:
            category (nark.Category): Category to be updated.

        Returns:
            None: If everything went ok.

        Raises:
            KeyError: If the ``Category`` can not be found by the backend.
            TypeError: If category passed is not an nark.Category instance.
            ValueError: If category passed does not have an pk.
        """
        raise NotImplementedError

    # ***

    def get(self, pk):
        """
        Get an ``Category`` by its primary key.

        Args:
            pk (int): Primary key of the ``Category`` to be fetched.

        Returns:
            nark.Category: ``Category`` with given primary key.

        Raises:
            KeyError: If no ``Category`` with this primary key can be found
                by the backend.
        """

        raise NotImplementedError

    # ***

    def get_by_name(self, name):
        """
        Look up a category by its name.

        Args:
            name (str): Unique name of the ``Category`` to we want to fetch.

        Returns:
            nark.Category: ``Category`` with given name.

        Raises:
            KeyError: If no ``Category`` with this name was found by the backend.
        """
        raise NotImplementedError

    # ***

    def get_all(
        self,
        include_usage=True,
        count_results=False,
        since=None,
        until=None,
        deleted=False,
        hidden=False,
        key=None,
        search_term='',
        activity=False,
        sort_col='',
        sort_order='',
        limit='',
        offset='',
    ):
        """
        Return a list of all categories.

        Returns:
            list: List of ``Categories``, ordered by ``lower(name)``.
        """
        raise NotImplementedError




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'. If not, see <http://www.gnu.org/licenses/>.

"""Base class for Nark item instances."""

from __future__ import absolute_import, unicode_literals

from future.utils import python_2_unicode_compatible

__all__ = ('BaseItem', )


@python_2_unicode_compatible
class BaseItem(object):
    """Base class for all items."""

    def __init__(self, pk, name):
        self.pk = pk
        self.name = name

    def __repr__(self):
        parts = []
        for key in sorted(self.__dict__.keys()):
            parts.append(
                "{key}={val}".format(key=key, val=repr(getattr(self, key)))
            )
        repred = "{cls}({parts})".format(
            cls=self.__class__.__name__, parts=', '.join(parts),
        )
        return repred

    @property
    def unstored(self):
        return (not self.pk) or (self.pk < 0)




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines logging related function."""

from __future__ import absolute_import, unicode_literals

import logging

from ansi_escape_room import attr, fg


def formatter_basic(color=False):
    if not color:
        return formatter_basic_plain()
    return formatter_basic_color()


def formatter_basic_plain():
    formatter = logging.Formatter(
        '[%(levelname)s] '
        '%(asctime)s '
        '%(name)s '
        '%(funcName)s: '
        '%(message)s'
    )
    return formatter


def formatter_basic_color():
    formatter = logging.Formatter(
        '{grey_54}[{underlined}{magenta}%(levelname)s{reset}{grey_54}]{reset} '
        '{yellow}%(asctime)s{reset} '
        '{light_blue}%(name)s '
        '%(funcName)s{reset}: '
        '{bold}{green}%(message)s{reset}'.format(
            grey_54=fg('grey_54'),
            underlined=attr('underlined'),
            magenta=fg('magenta'),
            reset=attr('reset'),
            yellow=fg('yellow'),
            light_blue=fg('light_blue'),
            bold=attr('bold'),
            green=fg('green'),
        )
    )
    return formatter


def resolve_log_level(level):
    error = False
    try:
        try:
            log_level = int(level)
        except ValueError:
            log_level = logging.getLevelName(level)
    except KeyError:
        error = True
        log_level = logging.WARNING
    return log_level, error


def set_logger_level(logger_name, logger_log_level):
    logger = logging.getLogger(logger_name)
    logger.addHandler(logging.NullHandler())

    # (lb): BIZARRE: On a 14.04 machine, parent.handlers has StreamHandler
    #   in it, so it prints to console. This does not happen on a 16.04
    #   machine I also use. And I cannot determine the reason (both
    #   machines use a virtualenv configured exactly the same way, and
    #   the Python version is merely off by one PATCH).
    logger.parent.handlers = []
    logger.parent.addHandler(logging.NullHandler())

    log_level, warn_name = resolve_log_level(logger_log_level)

    try:
        logger.setLevel(int(log_level))
    except ValueError:
        warn_name = True
        logger.setLevel(logging.WARNING)

    if warn_name:
        logger.warning(
            _('Unknown log_level specified for ‘{}’: {}')
            .format(logger_name, logger_log_level)
        )

    return logger


def setup_handler(handler, formatter, *loggers):
    handler.setFormatter(formatter)
    for logger in loggers:
        logger.addHandler(handler)




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

import sys

import ansi_escape_room

__all__ = (
    'disable_colors',
    'enable_colors',
    'coloring',
    'set_coloring',
    'fg',
    'bg',
    'attr',
    'colorize',
    # Private:
    #  'map_color'
)


# (lb): Retrieve pointer to module object instance, so functions can set
# (otherwise function scope creates local variable). (I'll admit this feels a
# little weird, but we need to gait access to a package -- colored -- and it's
# either this; or use a module-level dictionary variable that module function
# can set; or make a Singleton class. Or fork colored and move this there.)
this = sys.modules[__name__]

this.ENABLE_COLORS = True


def disable_colors():
    this.ENABLE_COLORS = False


def enable_colors():
    this.ENABLE_COLORS = True


def coloring():
    return this.ENABLE_COLORS


def set_coloring(new_coloring):
    was_coloring = this.ENABLE_COLORS
    this.ENABLE_COLORS = new_coloring
    return was_coloring


def fg(color):
    if not coloring():
        return ''
    return ansi_escape_room.fg(map_color(color))


def bg(color):
    if not coloring():
        return ''
    return ansi_escape_room.bg(map_color(color))


def attr(color):
    if not coloring():
        return ''
    return ansi_escape_room.attr(map_color(color))


def colorize(text, color, *args):
    if not coloring():
        return text
    more_attrs = ''.join([ansi_escape_room.attr(attr) for attr in args])
    return '{}{}{}{}'.format(
        ansi_escape_room.fg(color),
        more_attrs,
        text,
        ansi_escape_room.attr('reset'),
    )


def map_color(color):
    # FIXME/2018-06-08: (lb): Need a way to easily change palette.
    # Should at least have two profiles, one for black on white; and t'other.
    # Search all uses of fg and bg, and maybe even map attr?
    return color




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

"""Fixtures needed to test helper submodule."""

from __future__ import absolute_import, unicode_literals

import codecs
import datetime
import os

import fauxfactory
import pytest
from configparser import ConfigParser
from nark.helpers import app_config
from nark.helpers.app_dirs import NarkAppDirs
from six import text_type


@pytest.fixture
def filename():
    """Provide a filename string."""
    return fauxfactory.gen_utf8()


@pytest.fixture
def filepath(tmpdir, filename):
    """Provide a fully qualified pathame within our tmp-dir."""
    return os.path.join(tmpdir.strpath, filename)


@pytest.fixture
def appdirs(mocker, tmpdir):
    """Provide mocked version specific user dirs using a tmpdir."""
    def ensure_directory_exists(directory):
        if not os.path.lexists(directory):
            os.makedirs(directory)
        return directory

    NarkAppDirs.user_config_dir = ensure_directory_exists(os.path.join(
        tmpdir.mkdir('config').strpath, 'nark/'))
    NarkAppDirs.user_data_dir = ensure_directory_exists(os.path.join(
        tmpdir.mkdir('data').strpath, 'nark/'))
    NarkAppDirs.user_cache_dir = ensure_directory_exists(os.path.join(
        tmpdir.mkdir('cache').strpath, 'nark/'))
    NarkAppDirs.user_log_dir = ensure_directory_exists(os.path.join(
        tmpdir.mkdir('log').strpath, 'nark/'))
    return NarkAppDirs


@pytest.fixture
def backend_config(appdirs):
    """Provide generic backend config."""
    appdir = appdirs(app_config.DEFAULT_APP_NAME)
    return app_config.get_default_backend_config(appdir)


@pytest.fixture
def configparser_instance(request):
    """Provide a ``ConfigParser`` instance and its expected config dict."""
    config = ConfigParser()
    config.add_section('Backend')
    config.set('Backend', 'store', 'sqlalchemy')
    config.set('Backend', 'db_engine', 'sqlite')
    config.set('Backend', 'db_path', '/tmp/hamster.db')
    config.set('Backend', 'db_host', 'www.example.com')
    config.set('Backend', 'db_port', '22')
    config.set('Backend', 'db_name', 'hamster')
    config.set('Backend', 'db_user', 'hamster')
    config.set('Backend', 'db_password', 'hamster')
    config.set('Backend', 'day_start', '05:00:00')
    config.set('Backend', 'fact_min_delta', '60')
    config.set('Backend', 'lib_log_level', 'WARNING')
    config.set('Backend', 'sql_log_level', 'WARNING')
    # MAYBE: (lb): Consider fiddling with day_start and fact_min_delta
    # in specific tests and leaving them set to factory defaults here.
    #   config.set('Backend', 'day_start', '')
    #   config.set('Backend', 'fact_min_delta', '0')
    # Also consider the other settings not being set here.
    #   config.set('Backend', 'allow_momentaneous', 'False')
    #   config.set('Backend', 'tz_aware', 'False')
    #   config.set('Backend', 'default_tzinfo', '')

    expectation = {
        'store': text_type('sqlalchemy'),
        'db_engine': text_type('sqlite'),
        'db_path': text_type('/tmp/hamster.db'),
        'db_host': text_type('www.example.com'),
        'db_port': 22,
        'db_name': text_type('hamster'),
        'db_user': text_type('hamster'),
        'db_password': text_type('hamster'),
        'allow_momentaneous': False,
        'day_start': datetime.datetime.strptime('05:00:00', '%H:%M:%S').time(),
        'fact_min_delta': 60,
        # MAYBE: (lb): Consider fiddling with day_start and fact_min_delta
        # in specific tests and leaving them set to factory defaults here:
        #   'day_start': '',
        #   'fact_min_delta': 0,
        'lib_log_level': text_type('WARNING'),
        'sql_log_level': text_type('WARNING'),
        'tz_aware': False,
        'default_tzinfo': '',
    }

    return config, expectation


@pytest.fixture
def config_instance(request):
    """A dummy instance of ``ConfigParser``."""
    return ConfigParser()


@pytest.fixture
def config_file(backend_config, appdirs):
    """Provide a config file stored under our fake config dir."""
    config_conf = os.path.join(appdirs.user_config_dir, 'config.conf')
    with codecs.open(config_conf, 'w', encoding='utf-8') as fobj:
        app_config.backend_config_to_configparser(backend_config).write(fobj)
        config_instance.write(fobj)


@pytest.fixture(params=[
    ('foobar', {
        'timeinfo': None,
        'activity': 'foobar',
        'category': None,
        'description': None,
    }),
    ('11:00 12:00 foo@bar', {
        'timeinfo': '11:00',
        'activity': '12:00 foo',
        'category': 'bar',
        'description': None,
    }),
    ('rumpelratz foo@bar', {
        'timeinfo': None,
        'start': None,
        'end': None,
        'activity': 'rumpelratz foo',
        'category': 'bar',
        'description': None,
    }),
    ('foo@bar', {
        'timeinfo': '',
        'activity': 'foo',
        'category': 'bar',
        'description': None,
    }),
    ('foo@bar, palimpalum', {
        'timeinfo': None,
        'activity': 'foo',
        'category': 'bar',
        'description': 'palimpalum',
    }),
    ('12:00 foo@bar, palimpalum', {
        'timeinfo': '12:00',
        'activity': 'foo',
        'category': 'bar',
        'description': 'palimpalum',
    }),
    ('12:00 - 14:14 foo@bar, palimpalum', {
        'timeinfo': '12:00 to 14:14',
        'activity': 'foo',
        'category': 'bar',
        'description': 'palimpalum',
    }),
    # Missing whitespace around ``-`` will prevent timeinfo from being parsed.
    ('12:00-14:14 foo@bar, palimpalum', {
        'timeinfo': '',
        'activity': '12:00-14:14 foo',
        'category': 'bar',
        'description': 'palimpalum',
    }),
])
def raw_fact_parametrized(request):
    """Provide a variety of raw facts as well as a dict of its proper components."""
    return request.param




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

"""
Packaging instruction for setup tools.

Refs:

  https://setuptools.readthedocs.io/

  https://packaging.python.org/en/latest/distributing.html

  https://github.com/pypa/sampleproject
"""

from setuptools import find_packages, setup

# *** Package requirements.

requirements = [
    # "Very simple Python library for color and formatting in terminal."
    # Forked (for italic "support") to:
    #  https://github.com/hotoffthehamster/ansi-escape-room
    # Forked from:
    #  https://gitlab.com/dslackw/colored
    'ansi-escape-room',
    # Platform-specific directory magic.
    #  https://github.com/ActiveState/appdirs
    'appdirs',
    # Basic INI parser.
    #  https://pypi.org/project/configparser/
    #  https://docs.python.org/3/library/configparser.html
    'configparser >= 3.5.0b2',
    # https://github.com/scrapinghub/dateparser
    'dateparser',
    # Py2/3 support shim. (Higher-level than `six`.)
    #  https://pypi.org/project/future/
    #  https://python-future.org/
    'future',
    # Elapsed timedelta formatter, e.g., "1.25 days".
    'human-friendly_pedantic-timedelta >= 0.0.5',
    # https://github.com/collective/icalendar
    'icalendar',
    # https://bitbucket.org/micktwomey/pyiso8601
    'iso8601',
    # https://github.com/mnmelo/lazy_import
    'lazy_import',
    # Daylight saving time-aware timezone library.
    #  https://pythonhosted.org/pytz/
    'pytz',
    # For testing with dateparser,
    #   https://bitbucket.org/mrabarnett/mrab-regex
    #   https://pypi.org/project/regex/
    # FIXME/2019-02-19 18:13: Latest regex is broken.
    #   https://pypi.org/project/regex/2019.02.20/
    #   https://bitbucket.org/mrabarnett/mrab-regex/commits/
    #       5d8b8bb2b64b696cdbaa7bdc0dce4b462d311134#Lregex_3/regex.pyF400
    # Because of recent change to import line (was made self-referential):
    #      .tox/py37/lib/python3.7/site-packages/regex.py:400: in <module>
    #          import regex._regex_core as _regex_core
    #      E   ModuleNotFoundError: No module named 'regex._regex_core';
    #       'regex' is not a package
    'regex == 2019.02.18',
    # https://pythonhosted.org/six/
    'six',
    # https://www.sqlalchemy.org/
    'sqlalchemy',
    # Database gooser/versioner.
    #  https://pypi.org/project/sqlalchemy-migrate/
    #  https://sqlalchemy-migrate.readthedocs.io/en/latest/
    # 2019-02-21: (lb): Forked again! Package alt. that accepts static config.
    'sqlalchemy-migrate-dob >= 0.12.1',
    # https://github.com/regebro/tzlocal
    'tzlocal',
]

# *** Minimal setup() function -- Prefer using config where possible.

# (lb): All settings are in setup.cfg, except identifying packages.
# (We could find-packages from within setup.cfg, but it's convoluted.)

setup(
    install_requires=requirements,
    packages=find_packages(),
    # Tell setuptools to determine the version
    # from the latest SCM (git) version tags.
    #
    # Without the following two lines, e.g.,
    #   $ python setup.py --version
    #   3.0.0a31
    # But with 'em, e.g.,
    #   $ python setup.py --version
    #   3.0.0a32.dev3+g6f93d8c.d20190221
    # Or, if the latest commit is tagged,
    # and your working directory is clean,
    # then the version reported (and, e.g.,
    # used on make-dist) will be from tag.
    # Ref:
    #   https://github.com/pypa/setuptools_scm
    setup_requires=['setuptools_scm'],
    use_scm_version=True,
)




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

""""""

from __future__ import absolute_import, unicode_literals

from datetime import datetime

__all__ = (
    'antecedent_fact',
    'subsequent_fact',
)


def antecedent_fact(facts_mgr, new_facts, now_time):
    for fact in new_facts:
        # 2019-01-19: This isn't quite right. We could use earliest
        # clock time if we find first datetime, then look again for
        # clock time....
        if (
            (fact.start and isinstance(fact.start, datetime))
            or (fact.end and isinstance(fact.end, datetime))
        ):
            return facts_mgr.antecedent(fact=fact)
    return facts_mgr.antecedent(ref_time=now_time)


def subsequent_fact(facts_mgr, new_facts):
    for fact in reversed(new_facts):
        if (
            (fact.end and isinstance(fact.end, datetime))
            or (fact.start and isinstance(fact.start, datetime))
        ):
            return facts_mgr.subsequent(fact=fact)
    return None




# -*- coding: utf-8 -*-

# This file is part of 'nark'.
#
# 'nark' is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# 'nark' is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with 'nark'.  If not, see <http://www.gnu.org/licenses/>.

"""
Provide functions that provide common config related functionality.

This module provide easy to use convenience functions to handle common
configuration related tasks. Clients can use those to provide consistent
behaviour and focus on their specific requirements instead.

The easiest way to make use of those helpers is to call ``load_config_file``
(which will also handle creating a new one if none exists) and
``write_config_file``.

Clients may use ``backend_config_to_configparser`` and its counterpart
``configparser_to_backend_config`` to delegate conversion between a backend
config dict and a ``ConfigParser`` instance.

Note:
    Backend config key/value information:
        store: A ``string`` indicating which store to use.
            See: ``nark.REGISTERED_BACKENDS``

        db_engine: ``string`` indicating which db-engine to use.
            Options depend on store choice.

        db_path: ``string`` indicating where to save the db file if the
            selected db option saves to disk. Depends on store/engine choice.

        db_host: ``string`` indicating the host of the db server.
            Depends on store/engine choice.

        db_port: ``int`` indicating the port of the db server.
            Depends on store/engine choice.

        db_name: ``string`` indicating the db-name. Depends on store/engine choice.

        db_user: ``string`` indicating the username to access the db server.
            Depends on store/engine choice.

        db_password: ``string`` indicating the password to access the db server.
            Depends on store/engine choice.

        allow_momentaneous: ``bool`` indicting if 0-length duration Facts allowed,
            e.g., start == end. Here's a little vocabulary lesson:
                Fugacious: "lasting a short time"
                Evanescent: "tending to vanish like vapor"
                Momentaneous: "characterizing action begun, terminated in an instant"
                See also: Vacant, Fleeting, Transient.

        day_start: ``datetime.time`` cna be used to specify default start time.
            (Weird Legacy Hamster feature; now disabled by default.)

        fact_min_delta: ``int`` specifying minimal fact duration. Facts shorter
            than this will be rejected.

        lib_log_level: ``string`` indicating the nark library logging logger log level.

        sql_log_level: ``string`` indicating the SQLAlchemy logging logger log level.

        tz_aware: ``bool`` indicating if datetimes are relative UTC or not.

        default_tzinfo: ``string`` default Timezone to use when storing time.

    Please also note that a backend *config dict* does except ``None`` / ``empty``
    values, its ``ConfigParser`` representation does not include those however!
"""

from __future__ import absolute_import, unicode_literals

import datetime
import os

from configparser import ConfigParser, NoOptionError
from six import string_types, text_type

from ..control import REGISTERED_BACKENDS
from .app_dirs import NarkAppDirs

# NOTE: (lb): This file is unused by nark/dob.
#       It's used by hamster-gtk, which is broke
#       against nark because of so much refactoring.
# FIXME: Fix this file to work with hamster-gtk again?
#        (I'm not sure how much I care about the windowed app;
#         I'm pretty happy with just the CLI....)


DEFAULT_APP_NAME = 'nark'
DEFAULT_APPDIRS = NarkAppDirs(DEFAULT_APP_NAME)
DEFAULT_CONFIG_FILENAME = '{}.conf'.format(DEFAULT_APPDIRS.appname)


def get_config_path(appdirs=DEFAULT_APPDIRS, file_name=DEFAULT_CONFIG_FILENAME):
    """
    Return the path where the config file is stored.

    Args:
        app_name (text_type, optional): Name of the application, defaults to
        ``'projecthamster``. Allows you to use your own application specific
        namespace if you wish.
        file_name (text_type, optional): Name of the config file. Defaults to
        ``config.conf``.

    Returns:
        str: Fully qualified path (dir & filename) where we expect the config file.
    """
    if isinstance(appdirs, string_types):
        appdirs = NarkAppDirs(appdirs)
    return os.path.join(appdirs.user_config_dir, file_name)


def write_config_file(
    config_instance,
    appdirs=DEFAULT_APPDIRS,
    file_name=DEFAULT_CONFIG_FILENAME,
):
    """
    Write a ConfigParser instance to file at the correct location.

    Args:
        config_instance: Config instance to safe to file.
        appdirs (NarkAppDirs, optional): ``NarkAppDirs`` instance storing
            app/user specific path information.
        file_name (text_type, optional): Name of the config file. Defaults to
        ``DEFAULT_CONFIG_FILENAME``.

    Returns:
        ConfigParser: Instance written to file.
    """

    path = get_config_path(appdirs, file_name)
    with open(path, 'w') as fobj:
        config_instance.write(fobj)
    return config_instance


def load_config_file(
    appdirs=DEFAULT_APPDIRS,
    file_name=DEFAULT_CONFIG_FILENAME,
    fallback_config_instance=None,
):
    """
    Retrieve config information from file at default location.

    If no config file is found a new one will be created either with
        ``fallback_config_instance`` as content or if none is provided
        with the result of ``get_default_backend_config``.

    Args:
        appdirs (NarkAppDirs, optional): ``NarkAppDirs`` instance
            storing app/user specific path information.
        file_name (text_type, optional): Name of the config file.
            Defaults to ``DEFAULT_CONFIG_FILENAME``.
        fallback_config_instance (ConfigParser): Backend config that is
            to be used to populate the config file that is created if no
            pre-existing one can be found.

    Returns:
        ConfigParser: Config loaded from file, either from the the
            pre-existing config file or the one created with fallback values.
    """
    if not fallback_config_instance:
        fallback_config_instance = backend_config_to_configparser(
            get_default_backend_config(appdirs)
        )

    config = ConfigParser()
    path = get_config_path(appdirs, file_name)
    if not config.read(path):
        config = write_config_file(
            fallback_config_instance, appdirs=appdirs, file_name=file_name
        )
    return config


def get_default_backend_config(appdirs):
    """
    Return a default config dictionary.

    Args:
        appdirs (NarkAppDirs): ``NarkAppDirs`` instance encapsulating
            the apps details.

    Returns:
        dict: Dictionary with a default configuration.

    Note:
        Those defaults are independent of the particular config-store.
    """
    return {
        'store': 'sqlalchemy',
        'db_engine': 'sqlite',
        'db_path': os.path.join(
            appdirs.user_data_dir,
            '{}.sqlite'.format(appdirs.appname),
        ),
        # Skipping:
        #  'db_host': '',
        #  'db_port': '',
        #  'db_name': '',
        #  'db_user': '',
        #  'db_password': '',
        'allow_momentaneous': False,
        'day_start': '',
        'fact_min_delta': 0,
        'lib_log_level': 'WARNING',
        'sql_log_level': 'WARNING',
        'tz_aware': False,
        'default_tzinfo': '',
    }


# [TODO]
# Provide better error handling
def backend_config_to_configparser(config):
    """
    Return a ConfigParser instance representing a given backend config dictionary.

    Args:
        config (dict): Dictionary of config key/value pairs.

    Returns:
        ConfigParser: ConfigParser instance representing config.

    Note:
        We do not provide *any* validation about mandatory values whatsoever.
    """
    # NOTE: config.get(k) returns None on key miss; but config[k] would KeyError.
    def get_store():
        return config.get('store')

    def get_db_engine():
        return text_type(config.get('db_engine'))

    def get_db_path():
        return text_type(config.get('db_path'))

    def get_db_host():
        return text_type(config.get('db_host'))

    def get_db_port():
        return text_type(config.get('db_port'))

    def get_db_name():
        return text_type(config.get('db_name'))

    def get_db_user():
        return text_type(config.get('db_user'))

    def get_db_password():
        return text_type(config.get('db_password'))

    def get_allow_momentaneous():
        return text_type(bool(config.get('allow_momentaneous')))

    def get_day_start():
        day_start = config.get('day_start')
        if not day_start:
            return text_type('')
        return day_start.strftime('%H:%M:%S')

    def get_fact_min_delta():
        return text_type(config.get('fact_min_delta'))

    def get_lib_log_level():
        return text_type(config.get('lib_log_level'))

    def get_sql_log_level():
        return text_type(config.get('sql_log_level'))

    def get_tz_aware():
        # Convert to a string, e.g., True => 'True'.
        return text_type(bool(config.get('tz_aware')))

    def get_default_tzinfo():
        return text_type(config.get('default_tzinfo'))

    cp_instance = ConfigParser()
    cp_instance.add_section('Backend')
    cp_instance.set('Backend', 'store', get_store())
    cp_instance.set('Backend', 'db_engine', get_db_engine())
    cp_instance.set('Backend', 'db_path', get_db_path())
    cp_instance.set('Backend', 'db_host', get_db_host())
    cp_instance.set('Backend', 'db_port', get_db_port())
    cp_instance.set('Backend', 'db_name', get_db_name())
    cp_instance.set('Backend', 'db_user', get_db_user())
    cp_instance.set('Backend', 'db_password', get_db_password())
    cp_instance.set('Backend', 'allow_momentaneous', get_allow_momentaneous())
    cp_instance.set('Backend', 'day_start', get_day_start())
    cp_instance.set('Backend', 'fact_min_delta', get_fact_min_delta())
    cp_instance.set('Backend', 'lib_log_level', get_lib_log_level())
    cp_instance.set('Backend', 'sql_log_level', get_sql_log_level())
    cp_instance.set('Backend', 'tz_aware', get_tz_aware())
    cp_instance.set('Backend', 'default_tzinfo', get_default_tzinfo())

    return cp_instance


# [TODO]
# Provide better error handling
# Provide validation! For this it would probably be enough to validate a config
# dict. We do not actually need to validate a CP-instance but just its resulting
# dict.
def configparser_to_backend_config(cp_instance):
    """
    Return a config dict generated from a configparser instance.

    This functions main purpose is to ensure config dict values are properly typed.

    Note:
        This can be used with any ``ConfigParser`` backend instance not just
          the default one in order to extract its config.
        If a key is not found in ``cp_instance`` the resulting dict will have
          ``None`` assigned to this dict key.
    """
    def cp_instance_get(section, keyname, default=None):
        try:
            return cp_instance.get(section, keyname)
        except NoOptionError:
            return default

    def cp_instance_getboolean(section, keyname, default=None):
        try:
            return cp_instance.getboolean(section, keyname)
        except NoOptionError:
            return default

    def cp_instance_getint(section, keyname, default=None):
        try:
            return cp_instance.getint(section, keyname)
        except NoOptionError:
            return default

    # ***

    def get_store():
        # (lb): Here's a comment from hamster-lib:
        #     "This should be deligated to a dedicated validation function!"
        #   Though I'm not sure what's the ask. Should we
        #   check more than `store in REGISTERED_BACKENDS.keys`?
        # MAYBE: (lb): Use default 'sqlalchemy' if store not set?
        store = cp_instance_get('Backend', 'store')
        if store not in REGISTERED_BACKENDS.keys():
            raise ValueError(_("Unrecognized store option."))
        return store

    def get_db_engine():
        # (lb): Use default 'sqlite' if db_engine not set?
        return text_type(cp_instance_get('Backend', 'db_engine'))

    def get_db_path():
        return text_type(cp_instance_get('Backend', 'db_path'))

    def get_db_host():
        return text_type(cp_instance_get('Backend', 'db_host'))

    def get_db_port():
        return cp_instance_getint('Backend', 'db_port')

    def get_db_name():
        return text_type(cp_instance_get('Backend', 'db_name'))

    def get_db_user():
        return text_type(cp_instance_get('Backend', 'db_user'))

    def get_db_password():
        return text_type(cp_instance_get('Backend', 'db_password'))

    def get_allow_momentaneous():
        return cp_instance_getboolean('Backend', 'allow_momentaneous', False)

    def get_day_start():
        day_start = cp_instance_get('Backend', 'day_start')
        if not day_start:
            return ''
        try:
            day_start = datetime.datetime.strptime(day_start, '%H:%M:%S').time()
        except ValueError:
            raise ValueError(_(
                "We encountered an error when parsing config's 'day_start' value!"
                " Aborting ..."
            ))
        return day_start

    def get_fact_min_delta():
        return cp_instance_getint('Backend', 'fact_min_delta')

    def get_lib_log_level():
        return text_type(cp_instance_get('Backend', 'lib_log_level'))

    def get_sql_log_level():
        return text_type(cp_instance_get('Backend', 'sql_log_level'))

    def get_tz_aware():
        return cp_instance_getboolean('Backend', 'tz_aware', False)

    def get_default_tzinfo():
        return text_type(cp_instance_get('Backend', 'default_tzinfo', ''))

    result = {
        'store': get_store(),
        'db_engine': get_db_engine(),
        'db_path': get_db_path(),
        'db_host': get_db_host(),
        'db_port': get_db_port(),
        'db_name': get_db_name(),
        'db_user': get_db_user(),
        'db_password': get_db_password(),
        'allow_momentaneous': get_allow_momentaneous(),
        'day_start': get_day_start(),
        'fact_min_delta': get_fact_min_delta(),
        'lib_log_level': get_lib_log_level(),
        'sql_log_level': get_sql_log_level(),
        'tz_aware': get_tz_aware(),
        'default_tzinfo': get_default_tzinfo(),
    }

    return result

